/*
   Randall Maas
   1999,

   This is from libvfs project.
   This modules allows IO to continue in the "background" on non-POSIX machines
   that support readv/writev.
*/

#include "IOQ.h"
#include "LDL.h"
#include <sys/time.h>
#include <errno.h>
//#define __USE_GNU
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>

// See if GNU goodness is defined, if not, we won't bother.  Our routine
// below will do the same thing, just not as quickly
#ifndef TEMP_FAILURE_RETY
# define TEMP_FAILURE_RETRY(x) x
#endif

static pid_t my_pid=-1;
void FD_bkgnd(int FD)
{
   /* Set it up to be nonblocking and generate interrupts */
   fcntl (FD, F_SETFL, O_NONBLOCK | fcntl(FD, F_GETFL));
   fcntl (FD, F_SETFL, O_ASYNC    | fcntl(FD, F_GETFL));
   if (my_pid < 0) my_pid = getpid();
   fcntl (FD, F_SETOWN, my_pid);
}

#if IO_NPRIO > 1
# define FD2CQ(Fd) ComplQ[FD_Priority(Fd)]
# define FD2FQ(Fd) FailedQ[FD_Priority(Fd)]
#else
# define FD2CQ(Fd) ComplQ[0]
# define FD2FQ(Fd) FailedQ[0]
#endif

/*
   @node Submitting IO, IO
   @subheading Submitting background IO
   When you have a some IO to do, and would like it to occur in the background,
   you will need to fill out a IO request structure (@code{__IO_req_t}):

   @example
    __IO_req_t Req;
    struct iovec IOs[2];
    char BufA[256],BufB[512];

    Req.Op = 0;  // A read operation.  1 would be a write operation
    Req.FD = fd; // The file descriptor to read from
    Req.Count=2; // We will read into two buffers: A and B
    Req.IOReq=IOs;
    IOs[0].iov_base=(__ptr_t)BufA; // Read 256 bytes into buffer A
    IOs[0].iov_len = sizeof(BufA);
    IOs[1].iov_base=(__ptr_t)BufB; // Then read 512 bytes into buffer B
    IOs[1].iov_len = sizeof(BufB);

    LDL_append(&Req, SubmitQ);
   @end example

   @emph{Note 1:} The IO request structure is retained in the queue until it is
   completed.

   @emph{Note 2:} The IO channel may be ready for IO, so your program should
   plan on calling @code{__IOQ_run} sometime soon.

   @subheading Handling completed IO
    Any IO request that is placed into the completion queue or the failed queue
    is done.  It is up to your program to manage those resources as
    appropriate.

    The @var{Status} field of each element is used to indicate:
    @table @asis
    @item > 0
     The number of bytes read

    @item == 0
     No data was read, the file descriptor has no more data (EOF).  It should
     be closed.

    @item < 0
     Significant error.  The value of @var{Status} the negative of the
     @var{errno}.
    @end table

   @node SIGIO, IO
   @subheading Setting up to recieve notification.

   On platforms that define @code{O_ASYNC}, your program can receive
   @code{SIGIO} notification when a file descriptor is ready to be read or
   written.  

   Files that are opened should use the @code{O_ASYNC} flag to prevent
   execution from being blocked from continuing further.

   For file descriptors and IO streams that are already open, these can be
   set via a call like @code{fcntl(fd, F_SETFL, O_ASYNC);}

   After the signal has been handled, the @code{__IOQ_run} routine should be
   called to retrieve all of the data from the available IO channels.

   If you would like to pause your program while waiting for any IO, use the
   @code{pause} system call.  Your program will continue after a signal has
   been received.
*/

int __IOQ_run (__IO_req_t*  SubmitQ, __IO_req_t** ComplQ,
	       __IO_req_t** FailedQ, fd_set* IOSet)
{
  /*
     Description
      This routine submits a series of IO operations to be performed on all
      IO channels that are available.  It does this in way to help prevent
      blocking.  This is designed to run either during an @emph{idle loop}, or
      after a @code{SIGIO} signal.

      Each IO item stored in @code{SubmitQ} is checked to see if the IO channel
      is ready, and submitted if it is.  If the IO channel was set up for
      non-blocking operations, the submittal is guaranteed not to block.  The
      IO is dequeued on two events:
      @enumerate
      @item The IO was successful.  The IO element is then placed into the
       completion queue (@var{ComplQ}).
      @item The IO received an unrecoverable error.  The IO element is then
       placed into the failed queue (@var{FailedQ} -- this may be the same as
       the completion queue).
      @end enumerate

      Otherwise, the IO is left in the @code{SubmitQ}.

      The @var{Status} field of each element is used to indicate:
      @table @asis
      @item > 0
       The number of bytes read

      @item == 0
       No data was read, the file descriptor has no more data (EOF).  It should
       be closed.

      @item < 0
       Significant error.  The value of @var{Status} the negative of the
       @var{errno}.
      @end table

     @subheading Non-blocking IO

      Files that are opened should be use the @code{O_NONBLOCK} or
      @code{O_NDELAY} flags to allow execution to continue even if there is no
      data avaible (or can be sent).

      For file descriptors and IO streams that are already open, these can be
      set via a call like @code{fcntl(fd, F_SETFL, O_NONBLOCK);}

      (@pxref{SIGIO} for information on how to recieve notification when IO can
      be done).

     Return Value
      The number of items serviced.
  */


  int DQ = 0; //The number of items dequeue'd
  __IO_req_t* IPtr;
  ssize_t Status;

  /* First, get a status set of all of the files */
  struct timeval TV={0}; //Set for immediate timeout
  fd_set Read_FDs, Write_FDs;

  memcpy(&Read_FDs, IOSet, sizeof(Read_FDs));
  memcpy(&Write_FDs,IOSet, sizeof(Write_FDs));
  select(FD_SETSIZE, &Read_FDs, &Write_FDs, NULL, &TV);

  LDL_foreach(IPtr,(__IO_req_t*) SubmitQ)
   {
    PAST_GO:
      if (!IPtr->Op) // Read
	{
	   /* Check to see if the descriptor is ready for reading */
	   if (!FD_ISSET(IPtr->FD, &Read_FDs)) continue;

	   /* We use a macro to handle interrupted stuff, and retry the IO */
	   Status =TEMP_FAILURE_RETRY(readv(IPtr->FD,IPtr->IOReq,IPtr->Count));
	}
       else
	{
	   /* Operation is Write */
	   /* Check to see if the descriptor is ready for writing */
	   if (!FD_ISSET(IPtr->FD, &Write_FDs)) continue;

	   /* We use a macro to handle interrupted stuff, and retry the IO */
	   Status=TEMP_FAILURE_RETRY(writev(IPtr->FD,IPtr->IOReq,IPtr->Count));
	}

       if (Status >= 0)
	 {
	    /* We have transferred a valid amount of data... remove the item
	      from the to-do list, and put it into the completion queue */
	    __IO_req_t* JPtr = IPtr->q_forw;
	    remque(IPtr);
	    IPtr->Status = Status; // Indicates the size actually read
	    DQ++;

	    LDL_append(IPtr, FD2CQ(IPtr->FD));
	    if (!JPtr || JPtr == (__IO_req_t*)SubmitQ) break;
	    IPtr=JPtr;
	    goto PAST_GO;
	 }

       /* Handle errors! */
       switch (errno)
	 {
	    case EINTR:         // Interrupt came in, and not handle above...?!
# if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
	    case EWOULDBLOCK:
# endif
# ifdef EAGAIN
	    case EAGAIN:
# endif
	      /* Squelch any further IO on this channel for this cycle... */
	      FD_CLR(IPtr->FD, &Write_FDs);
	      FD_CLR(IPtr->FD, &Read_FDs);
	      break; // IO is not really ready for us yet..

	    case EIO:   // Hardware error
	    case EBADF: // Bogus file descriptor
	    case EFBIG: // Write takes up way too much space on disk
	    case ENOSPC:// Disk is full
	    case EPIPE: // Pipe is not open on other end!
	    case EFAULT: // Something points to bogus memory
	    case EINVAL: //An invalid argument
	    case EISDIR: // FD refers to a directory
	    case EOPNOTSUPP: // FD doesn't support operation

	    default:    // Not supposed to happen!
	     {
	      /* Basically we drop this into the error bin... */
	      __IO_req_t* JPtr = IPtr->q_forw;
	      remque(IPtr);
	      IPtr->Status=-errno; // Indicate error
	      DQ++;
	      LDL_append(IPtr,FD2FQ(IPtr->FD));
	      if (!JPtr || JPtr == (__IO_req_t*) SubmitQ) return DQ;
	      IPtr=JPtr;
	      goto PAST_GO;
	     }
	 }
   }

  return DQ;
}
