/*

 gzilla

 Copyright 1997 Raph Levien <raph@acm.org>

 This code is free for commercial and non-commercial use,
 modification, and redistribution, as long as the source code release,
 startup screen, or product packaging includes this copyright notice.

 */

/* The gif decoder for gzilla. It is responsible for decoding GIF data
   in a bytesink and transferring it to an imgsink.

   Adapted by Raph Levien from giftopnm.c as found in the
   netpbm-1mar1994 release. The copyright notice for giftopnm.c
   follows: */

/* +-------------------------------------------------------------------+ */
/* | Copyright 1990, 1991, 1993, David Koblas.  (koblas@netcom.com)    | */
/* |   Permission to use, copy, modify, and distribute this software   | */
/* |   and its documentation for any purpose and without fee is hereby | */
/* |   granted, provided that the above copyright notice appear in all | */
/* |   copies and that both that copyright notice and this permission  | */
/* |   notice appear in supporting documentation.  This software is    | */
/* |   provided "as is" without express or implied warranty.           | */
/* +-------------------------------------------------------------------+ */


/* Notes 13 Oct 1997

   Today, just for the hell of it, I implemented a new decoder from
   scratch. It's oriented around pushing bytes, while the old decoder
   was based around reads which may suspend. There were basically
   three motivations.

   1. To increase the speed.

   2. To fix some bugs I had seen, most likely due to suspension.

   3. To make sure that the code had no buffer overruns or the like.

   4. So that the code could be released under a freer license.

   Let's see how we did on speed. I used a large image for testing
   (fvwm95-2.gif).

   The old decoder spent a total of about 1.04 seconds decoding the
   image. Another .58 seconds went into gzilla_image_line, almost
   entirely conversion from colormap to RGB.

   The new decoder spent a total of 0.46 seconds decoding the image.
   However, the time for gzilla_image_line went up to 1.01 seconds.
   Thus, even though the decoder seems to be about twice as fast,
   the net gain is pretty minimal. Could this be because of cache
   effects?

   Lessons learned: The first, which I keep learning over and over, is
   not to try to optimize too much. It doesn't work. Just keep things
   simple. 

   Second, it seems that the colormap to RGB conversion is really a
   significant part of the overall time. It's possible that going
   directly to 16 bits would help, but that's optimization again :)

*/

/* Todos:

 + Make sure to handle error cases gracefully (including aborting the
   imgsink, if necessary).

 */

#undef VERBOSE

#include <stdio.h>  /* for sprintf */
#include <string.h> /* for memcpy and memmove */
#include <malloc.h>

#include <gtk/gtk.h>
#include "gzillaimgsink.h"
#include "gzillacache.h"
#include "gzillaweb.h"
#include "MIME/MIME.h"


#define INTERLACE              0x40
#define LOCALCOLORMAP  0x80

#define LM_to_uint(a,b)                        ((((unsigned char)b)<<8)|((unsigned char)a))

#define GZILLA_GIF_CLEAN 0
#define GZILLA_GIF_DIRTY 1
#define GZILLA_GIF_SIZECHANGE 2

/* The private, internal state of the gif byte sink. */
#define        MAXCOLORMAPSIZE         256
#define        MAX_LWZ_BITS            12


typedef struct _GzillaGif {
  GzillaImgSink  *imgsink;

  gint            state;
  size_t          Start_Ofs;

  unsigned char   input_code_size;
  unsigned char   label;
  gboolean        interlace;
  gchar          *linebuf;
  gint            pass;

  gint            x, y;

  /* state for lwz_read_byte */
  int		code_size;
  int             fresh;

  /* The original GifScreen from giftopnm */
  unsigned int    Width;
  unsigned int    Height;
  size_t	  ColorMap_ofs;
  unsigned int    BitPixel;
  unsigned int    ColorResolution;
  unsigned int    Background;
  unsigned int    AspectRatio;

  /* Gif89 extensions */
  int             transparent;
  int             delayTime;
  int             inputFlag;
  int             disposal;

  /* state for the new push-oriented decoder */
  gint            packet_size;
  guint           window;
  gint            bits_in_window;
  gint            last_code;
  gint            line_index;
  unsigned char   **spill_lines;
  gint            num_spill_lines_max;
  gint            length[(1 << MAX_LWZ_BITS) + 1];
  gint            code_and_byte[(1 << MAX_LWZ_BITS) + 1];
} GzillaGif;

/* Some invariants:

   last_code <= code_mask

   code_and_byte is stored packed: (code << 8) | byte
   */


static void gzilla_gif_write      (GzillaGif *gif, __CacheFile_t*CPtr);
static void gzilla_gif_close      (GzillaGif *gif);

static gint process_bytes (GzillaGif *gif, const unsigned char *buf, gint bufsize, __CacheFile_t* FPtr);

static GzillaGif* gzilla_gif_new (GzillaImgSink *imgsink);
static void gzilla_gif_callback(int Op, void* B, __CacheFile_t* CPtr);


Gzw* gzImage_GIF(const char* Type,void*Ptr, __IOCallback_t* Call,
		void**Data)
{
  Gzw* child_gzw=NULL;
  GzillaWeb* web=Ptr;
  GzillaImgSink *imgsink;

  if (web->imgsink)
    imgsink = web->imgsink;
  else {
    imgsink = gzilla_image_new (0, 0, NULL, 0xd6d6d6);
    imgsink->gzw = (GzwImage*)gzw_image_new (GZW_IMAGE_RGB);
    child_gzw = (Gzw*)imgsink->gzw;
  }
  
  *Data = gzilla_gif_new (imgsink);
  *Call=(__IOCallback_t) gzilla_gif_callback;

  return child_gzw;
}

static GzillaGif*
gzilla_gif_new (GzillaImgSink *imgsink)
{
  GzillaGif *gif=malloc(sizeof(*gif));

  gif->imgsink = imgsink;
  imgsink->RefCount++;
  imgsink->Parent=&gif->imgsink;
  gif->state = 0;
  gif->Start_Ofs=0;
  gif->linebuf = NULL;
  gif->fresh = FALSE;
  gif->transparent = -1;
  gif->num_spill_lines_max = 0;
  gif->spill_lines = 0;
  gif->window=0;
  gif->packet_size=0;

#ifdef VERBOSE
  g_print ("new gif %x\n", gif);
#endif

  return gif;
}

void gzilla_gif_callback(int Op, void* B, __CacheFile_t* FPtr)
{
   if (Op) gzilla_gif_close(B);
   else gzilla_gif_write(B,FPtr);
}
  
static void gzilla_gif_write (GzillaGif *gif, __CacheFile_t* FPtr)
{
  char *buf=((char*)FPtr->Data)+gif->Start_Ofs;
  ssize_t bufsize=FPtr->Size-gif->Start_Ofs;
if (!FPtr->Data) return;
if (!gif->imgsink->Parent) abort();
#ifdef VERBOSE
  g_print ("gzilla_gif_write: %d bytes\n", num);
#endif

  /* Process the bytes in the input buffer. */
    if (FPtr->Size <= gif->Start_Ofs) return;
    {
      ssize_t bytes_consumed = process_bytes (gif, buf, bufsize,FPtr);
      if (bytes_consumed < 1) return;
      gif->Start_Ofs += bytes_consumed;
   }

#ifdef VERBOSE
  g_print ("exit gzilla_gif_write, bufsize=%d\n", bufsize);
#endif
}

static void
gzilla_gif_close (GzillaGif *gif)
{
  gint i;

#ifdef VERBOSE
  g_print ("destroy gif %x\n", gif);
#endif
  gzilla_imgsink_close(gif->imgsink);
  gif->imgsink=NULL;
  if (gif->linebuf != NULL)
    free (gif->linebuf);

  if (gif->spill_lines != NULL)
    {
      for (i = 0; i < gif->num_spill_lines_max; i++)
	free (gif->spill_lines[i]);
      free (gif->spill_lines);
    }

   free(gif);
}

/* Return TRUE when the extension is over */
static size_t do_extension (GzillaGif *gif, const unsigned char *pbuf,
	size_t pbufsize)
{
  gint size;

  if (pbufsize <= 0 || pbufsize < 1+*pbuf) return 0;
  size = *pbuf++;
  if (!size)
    {
       gif->state = 3;
       return 1;
    }

  if (gif->label != 0xf9)
    {
       return size+1;
    }

  /* Graphic Control Extension */
  gif->disposal  = (pbuf[0] >> 2) & 0x7;
  gif->inputFlag = (pbuf[0] >> 1) & 0x1;
  gif->delayTime = LM_to_uint (pbuf[1], pbuf[2]);
  if (pbuf[0] & 0x1)
    gif->transparent = pbuf[3];
  return size+1;
}

/* Here begins the new push-oriented decoder. It should be quite a bit
   faster than the old code, which was adapted to be asynchronous from
   David Koblas's original giftopnm code.

 */

static void
gzilla_gif_lwz_init (GzillaGif *gif)
{
  gif->num_spill_lines_max = 1;
  gif->spill_lines = g_malloc(sizeof(unsigned char *)* gif->num_spill_lines_max);
  gif->spill_lines[0] = g_malloc(gif->Width);
  gif->bits_in_window = 0;
  gif->last_code = (1 << gif->input_code_size) + 1;
  gif->code_size = gif->input_code_size + 1;
  gif->line_index = 0;
}

/* Send the image line to the imgsink, also handling the interlacing. */
static void
gzilla_gif_emit_line (GzillaGif *gif, const unsigned char *linebuf)
{
  if (gif->y < gif->Height)
  gzilla_imgsink_write (gif->imgsink, linebuf, 0, gif->y);
  if (gif->interlace) {
    switch (gif->pass) {
    case 0:
    case 1:
      gif->y += 8;
      break;
    case 2:
      gif->y += 4;
      break;
    case 3:
      gif->y += 2;
      break;
    }
    if (gif->y >= gif->Height) {
      gif->pass++;
      switch (gif->pass) {
      case 1:
	gif->y = 4;//MIN(4,gif->Height-1);
	break;
      case 2:
	gif->y = 2;//MIN(2,gif->Height-1);
	break;
      case 3:
	gif->y = 1;
	break;
      default:
	/* arriving here is an error in the input image. */
	gif->y = 0;
	break;
      }
    }
    
  } else {
    if (gif->y < gif->Height)
      gif->y++;
  }
}

/* Decode the packetized lwz bytes into the image sink.

   Return codes:

    0 = working
    1 = successful eof code (but does not consume zero size packet)
   -1 = error

   I apologize for the large size of this routine and the goto error
   construct - I almost _never_ do that. I offer the excuse of
   optimizing for speed. */

static gint
gzilla_gif_decode (GzillaGif *gif, const unsigned char *buf, gint *pbufsize)
{
  gint bufsize;
  gint packet_size;
  gint clear_code;
  guint window;
  gint bits_in_window;
  unsigned code;
  unsigned last_code;
  gint code_size;
  unsigned code_mask;
  gint width;
  unsigned char *linebuf= gif->linebuf;
  gint line_index;
  gint sequence_length;
  gint num_spill_lines=0;
  unsigned char *obuf=NULL;
  unsigned o_index;
  unsigned o_size;
  gint spill_line_index=0;
  unsigned code_and_byte;
  gint num_lines_emit;
  unsigned orig_code;
  unsigned char *last_byte_ptr=NULL;

  bufsize = *pbufsize;

  /* Want to get all inner loop state into local variables. */
  packet_size = gif->packet_size;
  window = gif->window;
  bits_in_window = gif->bits_in_window;
  last_code = gif->last_code;
  code_size = gif->code_size;
  code_mask = (1 << code_size) - 1;
  line_index = gif->line_index;

  clear_code = 1 << gif->input_code_size;
  width = gif->Width;

  while (bufsize > 0)
    {
      /* lwz_bytes is the number of remaining lwz bytes in the packet. */
      gint lwz_bytes = MIN (packet_size, bufsize);
      bufsize -= lwz_bytes;
      packet_size -= lwz_bytes;
      for (; lwz_bytes > 0; lwz_bytes--)
	{
	  /* printf ("%d ", *buf) would print the depacketized lwz stream. */
	  window = (window >> 8) | (*buf++ << 24);
	  bits_in_window += 8;
	  while (bits_in_window >= code_size)
	    {
	      code = (window >> (32 - bits_in_window)) & code_mask;
	      /* printf ("%d\n", code); would print the code stream. */
	      /*
	      printf ("cs = %d, last_code = %d, code = %d, biw = %d, w = %x\n",
		      code_size, last_code, code, bits_in_window, window);
		      */
#ifdef VERBOSE
	      printf ("last_code = %d, code = %d, ",
		      last_code, code);
#endif
	      bits_in_window -= code_size;
	      if (code < clear_code)
		{
		  /* a literal code. */
#ifdef VERBOSE
		  printf ("literal\n");
#endif
		  linebuf[line_index++] = code;
		  if (line_index >= width)
		    {
		      gzilla_gif_emit_line (gif, gif->linebuf);
		      line_index = 0;
		    }
		  gif->length[last_code + 1] = 2;
		  gif->code_and_byte[last_code + 1] = (code << 8);
		  gif->code_and_byte[last_code] |= code;
		  last_code++;
		  if ((last_code & code_mask) == 0)
		    {
		      if (last_code == (1 << MAX_LWZ_BITS))
			last_code--;
		      else
			{
			  code_size++;
			  code_mask = (1 << code_size) - 1;
			}
		    }
		}
	      else if (code >= clear_code + 2)
		{
		  /* a sequence code. */
		  if (code > last_code) goto error;
		  sequence_length = gif->length[code];
		  gif->length[last_code + 1] = sequence_length + 1;
		  gif->code_and_byte[last_code + 1] = (code << 8);
		  /* We're going to traverse the sequence backwards. Thus,
		     we need to allocate spill lines if the sequence won't
		     fit entirely within the present scan line. */
		  if (line_index + sequence_length <= width)
		    {
		      num_spill_lines = 0;
		      obuf = linebuf;
		      o_index = line_index + sequence_length;
		      o_size = sequence_length - 1;
		    }
		  else
		    {
		      num_spill_lines = (line_index + sequence_length - 1) /
			width;
		      o_index = (line_index + sequence_length - 1) % width + 1;
		      if (num_spill_lines > gif->num_spill_lines_max)
			{
			  /* Allocate more spill lines. */
			  spill_line_index = gif->num_spill_lines_max;
			  do
			    gif->num_spill_lines_max <<= 1;
			  while (num_spill_lines > gif->num_spill_lines_max);
			  gif->spill_lines =
			    g_realloc (gif->spill_lines,
				       gif->num_spill_lines_max *
				       sizeof (unsigned char *));
			  for (; spill_line_index < gif->num_spill_lines_max;
			       spill_line_index++)
			    gif->spill_lines[spill_line_index] =
			      g_malloc(width);
			}
		      spill_line_index = num_spill_lines - 1;
		      obuf = gif->spill_lines[spill_line_index];
		      o_size = o_index;
		    }
		  line_index = o_index; /* for afterwards */

		  /* for fixing up later if last_code == code */
		  orig_code = code;
		  last_byte_ptr = obuf + o_index - 1;

		  /* spill lines are allocated, and we are clear to
		     write. This loop does not write the first byte of
		     the sequence, however (last byte traversed). */
		  while (sequence_length > 1)
		    {
		      sequence_length -= o_size;
		      /* Write o_size bytes to
			 obuf[o_index - o_size..o_index). */
		      for (; o_size > 0 && o_index>0; o_size--)
			{
#ifdef VERBOSE
			  printf ("%d ", gif->code_and_byte[code] & 255);
#endif
			  code_and_byte = gif->code_and_byte[code];
			  obuf[--o_index] = code_and_byte & 255;
			  code = code_and_byte >> 8;
			}
		      /* Prepare for writing to next line. */
		      if (o_index == 0)
			{
			  if (spill_line_index > 0)
			    {
			      spill_line_index--;
			      obuf = gif->spill_lines[spill_line_index];
			      o_size = width;
			    }
			  else
			    {
			      obuf = linebuf;
			      o_size = sequence_length - 1;
			    }
			  o_index = width;
			}
		    }
		  /* Ok, now we write the first byte of the sequence. */
		  /* We are sure that the code is literal. */
#ifdef VERBOSE
		  printf ("%d", code);
#endif
		  obuf[--o_index] = code;
		  gif->code_and_byte[last_code] |= code;

		  /* Fix up the output if the original code was last_code. */
		  if (orig_code == last_code) {
		    *last_byte_ptr = code;
#ifdef VERBOSE
		    printf (" fixed (%d)!", code);
#endif
		  }

#ifdef VERBOSE
		  printf ("\n");
#endif

		  /* Output any full lines. */
		  num_lines_emit = num_spill_lines;
		  if (line_index >= width)
		    {
		      num_lines_emit++;
		      line_index = 0;
		    }
		  if (num_lines_emit)
		    {
		      gzilla_gif_emit_line (gif, gif->linebuf);
		      for (spill_line_index = 0;
			   spill_line_index < num_lines_emit - 1;
			   spill_line_index++)
			gzilla_gif_emit_line
			  (gif, gif->spill_lines[spill_line_index]);
		    }

		  if (num_spill_lines > 0)
		    {
		      /* Swap the last spill line with the gif line, using
			 linebuf as the swap temporary. */
		      linebuf = gif->spill_lines[num_spill_lines - 1];
		      gif->spill_lines[num_spill_lines - 1] = gif->linebuf;
		      gif->linebuf = linebuf;
		    }

		  last_code++;
		  if ((last_code & code_mask) == 0)
		    {
		      if (last_code == (1 << MAX_LWZ_BITS))
			last_code--;
		      else
			{
			  code_size++;
			  code_mask = (1 << code_size) - 1;
			}
		    }
		}
	      else if (code == clear_code)
		{
		  /* clear code. */
#ifdef VERBOSE
		  printf ("clear\n");
#endif
		  code_size = gif->input_code_size + 1;
		  code_mask = (1 << code_size) - 1;
		  last_code = clear_code + 1;
		}
	      else
		{
		  /* end code. */
#ifdef VERBOSE
		  printf ("end\n");
#endif
		  //gzilla_imgsink_close (gif->imgsink);
		  //gif->imgsink = NULL;
		  *pbufsize = bufsize;
		  return 1;
		}
	    }
	}
      if (bufsize > 0)
	{
	  packet_size = *buf++;
	  bufsize--;
//	  if (packet_size == 0)  break; //goto error;
	}
    }

  gif->packet_size = packet_size;
  gif->window = window;
  gif->bits_in_window = bits_in_window;
  gif->last_code = last_code;
  gif->code_size = code_size;
  gif->line_index = line_index;
  *pbufsize=bufsize;
  return 0;

error:
  printf ("gzilla_gif_decode: error!\n");
  *pbufsize = bufsize;
  return -1;
}

static int check_sig(GzillaGif* gif, const char* ibuf, size_t Size)
{
    /* at beginning of file - read magic number */
    if (Size< 6) return 0;
    if (strncmp (ibuf, "GIF", 3) != 0) {
 fprintf(stderr,"bad sig\n");
      gif->state = 999;
      return 6;
    }
    if (strncmp (ibuf+ 3, "87a", 3) != 0 &&
	strncmp (ibuf+ 3, "89a", 3) != 0) {
 fprintf(stderr,"bad sig\n");
      gif->state = 999;
      return 6;
    }
    gif->state = 1;
   return 6;
}

static int get_descriptor(GzillaGif* gif, const unsigned char* buf,size_t Size)
{
    /* screen descriptor */
    if (Size< 7) return 0;
    gif->Width = LM_to_uint(buf[0],buf[1]);
    gif->Height = LM_to_uint(buf[2],buf[3]);
    gif->BitPixel = 2<<(buf[4]&0x07);
    gif->ColorResolution = (((buf[4]&0x70)>>3)+1);
    gif->Background      = buf[5];
    gif->AspectRatio     = buf[6];

    gif->linebuf = g_malloc(gif->Width);
    gzilla_imgsink_set_parms (gif->imgsink, gif->Width, gif->Height,
			      GZILLA_IMG_TYPE_INDEXED);

    if (buf[4] & LOCALCOLORMAP) {
      gif->state = 2;
    } else {
      gif->state = 3;
    }
  return 7;
}

/* Process some bytes from the input gif stream. It's a state machine. */
static gint process_bytes (GzillaGif *gif, const unsigned char *ibuf, gint bufsize, __CacheFile_t* FPtr) {
  char c;
  int tmp_bufsize=bufsize;
  size_t mysize;

  while(bufsize>0)
  switch (gif->state) {

  case 0: mysize=check_sig(gif,ibuf,tmp_bufsize);
	if (!mysize) return bufsize-tmp_bufsize;
	tmp_bufsize-= mysize;
	ibuf+= mysize;
	continue;
  case 1: mysize=get_descriptor(gif,ibuf,tmp_bufsize);
	if (!mysize) return bufsize-tmp_bufsize;
	tmp_bufsize-= mysize;
	ibuf+= mysize;
	continue;
  case 2:
    /* read colormap */
    if(gif->BitPixel*3 > tmp_bufsize)
	return bufsize - tmp_bufsize;
    gif->ColorMap_ofs = (unsigned long)ibuf- (unsigned long)FPtr->Data;
    gif->state = 3;
    tmp_bufsize -= gif->BitPixel*3;
    ibuf += gif->BitPixel*3;

  case 3:
    /* next character is one of:
       ';' = GIF terminator
       '!' = extension
       ',' = start character
       */
    if (tmp_bufsize < 1)
	return bufsize - tmp_bufsize;
    c = *ibuf++;
    tmp_bufsize--;
#ifdef VERBOSE
    g_print ("c = %c\n", c);
#endif
    if (c == ',') {
      gif->state = 4;
    } else if (c == '!') {
      gif->state = 7;
      continue;
    } else {
      gif->state = 999;
      return bufsize - tmp_bufsize;
    }

  case 4:
    /* right after the start character */
    if (tmp_bufsize < 9)
      return bufsize - tmp_bufsize;
    /* we should probably just check that the local stuff is consistent
       with the stuff at the header. For now, we punt... */
    gif->interlace = (ibuf[8] & INTERLACE) != 0;
    gif->pass = 0;
    tmp_bufsize -=9;
    if (ibuf[8] & LOCALCOLORMAP)
      {
	gif->state = 41;
        ibuf += 9;
      }
    else
      {
        ibuf += 9;
	gif->state = 5;
        continue;
      }
   
  case 41:
    /* read the local color map */
    if(gif->BitPixel*3 > tmp_bufsize)
      return bufsize - tmp_bufsize;
    gif->ColorMap_ofs = (unsigned long)ibuf- (unsigned long)FPtr->Data;
    gif->state = 5;
    tmp_bufsize-=gif->BitPixel*3;
    ibuf += gif->BitPixel*3;

  case 5:
    /* corresponds to ReadImage */
    if (tmp_bufsize < 1)
      return bufsize - tmp_bufsize;
    gif->input_code_size = *ibuf++;
    tmp_bufsize--;
    if (gif->input_code_size > 8)
      {
	gif->state = 999;
        return bufsize - tmp_bufsize;
      }
    gif->x = 0;
    gif->y = 0;
    gzilla_gif_lwz_init (gif);
    if (gif->imgsink)
      gzilla_imgsink_set_cmap (gif->imgsink,
			(unsigned char*)FPtr->Data+gif->ColorMap_ofs, 256,
			       gif->transparent);
    gif->state = 6;
//    if (tmp_bufsize <1) break;
    //Drop thru

  case 6:
    /* get an image byte */
    if (gzilla_gif_decode (gif, ibuf, &tmp_bufsize))
      gif->state = 999;
    return bufsize - tmp_bufsize;
    break;

  case 7:
    /* get the extension type */
    if (tmp_bufsize < 1)
      return bufsize - tmp_bufsize;
    gif->label=*ibuf++;
    gif->state = 8;
    tmp_bufsize--;
    // Drop thru to next state

  case 8:
    /* corresponds to DoExtension */
    mysize = do_extension (gif, ibuf, tmp_bufsize);
    tmp_bufsize -= mysize;
    ibuf += mysize;
    break;

  default:
    /* error - just consume all input */
    return bufsize - tmp_bufsize;
    //tmp_bufsize = 0;
  }
#ifdef VERBOSE
  printf ("process_bytes: initial state %d, final state %d, %d bytes consumed\n",
	  initial_state, gif->state, bufsize - tmp_bufsize);
#endif
  return bufsize - tmp_bufsize;
}
